﻿local self = sRaidFrames
local sRaidFrames = sRaidFrames
local L = LibStub("AceLocale-3.0"):GetLocale("sRaidFrames")
local Media = LibStub("LibSharedMedia-3.0")

function sRaidFrames:CreateIndicators(f)
	for _, k in pairs({"tl", "tr", "bl", "br"}) do
		local indicator = CreateFrame("Frame", nil, f)
		indicator:SetWidth(8)
		indicator:SetHeight(8)
		local backdrop = {
			bgFile = "Interface\\Addons\\sRaidFrames_Advanced\\indicator", tile = true, tileSize = 16,
			edgeFile = "Interface\\Addons\\sRaidFrames_Advanced\\indicator", edgeSize = 1,
			insets = {left = 1, right = 1, top = 1, bottom = 1},
		}
		indicator:SetBackdrop(backdrop)
		indicator:SetBackdropColor(0, 0, 0, 0)
		indicator:SetBackdropBorderColor(0, 0, 0, 0)
		indicator:Show()
		f["indicator_"..k] = indicator
	end
end

function sRaidFrames:PredictedHealthOnUpdate()
	if not self.PredictedHealthCache then
		self.PredictedHealthCache = {};
	end
	if not self.FramesByUnit then return end
	for unit in pairs(self.FramesByUnit) do
		local currhp = UnitHealth(unit)
		if currhp ~= self.PredictedHealthCache[unit] then
			self.PredictedHealthCache[unit] = currhp;
			self:UpdateUnitHealth(self, unit)
		end
	end
end

function sRaidFrames:CreateLayoutOptions(data, i)
	self.options.args.layouts.args[data.intname] = {
		name = data.name,
		type = "group",
		desc = L["Configure "]..data.name,
		args = {
			titlebar = {
				name = L["Title bar"],
				type = "group",
				desc = string.format(L["Configure this layout's %s bar"], L["title"]),
				order = 1,
				args = {
					height = {
						name = L["Height"],
						type = "input",
						desc = string.format(L["Modify the %s of this layout's %s bar"], L["height"], L["title"]),
						get = function() return tostring(self.opt.Layouts[i].title.height) or "16" end,
						set = function(info, value) self.opt.Layouts[i].title.height = tonumber(value); self:LayoutUpdate(i); end,
						order = 1,
					},
					width = {
						name = L["Width"],
						type = "input",
						desc = string.format(L["Modify the %s of this layout's %s bar"], L["width"], L["title"]),
						get = function() return tostring(self.opt.Layouts[i].title.width) or "80" end,
						set = function(info, value) self.opt.Layouts[i].title.width = tonumber(value); self:LayoutUpdate(i); end,
						order = 2,
					},
					offsets = {
						name = L["Offsets"],
						type = "group",
						desc = string.format(L["Modify the %s of this layout's %s bar"], L["offsets"], L["title"]),
						dialogInline = true,
						order = 3,
						args = {
							x = {
								name = "X",
								type = "input",
								desc = string.format(L["Modify the %s of this layout's %s bar"], "X "..L["offset"], L["title"]),
								get = function() return tostring(sRaidFrames.opt.Layouts[i].title.offsetx) or "0" end,
								set = function(info, value) sRaidFrames.opt.Layouts[i].title.offsetx = tonumber(value); self:LayoutUpdate(i); end,
								order = 1,
							},
							y = {
								name = "Y",
								type = "input",
								desc = string.format(L["Modify the %s of this layout's %s bar"], "Y "..L["offset"], L["title"]),
								get = function() return tostring(sRaidFrames.opt.Layouts[i].title.offsety) or "0" end,
								set = function(info, value) sRaidFrames.opt.Layouts[i].title.offsety = tonumber(value); self:LayoutUpdate(i); end,
								order = 1,
							},
						},
					},
				},
			},
			hpbar = {
				name = L["HP bar"],
				type = "group",
				desc = string.format(L["Configure this layout's %s bar"], L["HP"]),
				order = 2,
				args = {
					height = {
						name = L["Height"],
						type = "input",
						desc = string.format(L["Modify the %s of this layout's %s bar"], L["height"], L["HP"]),
						get = function() return tostring(sRaidFrames.opt.Layouts[i].hpbar.height) or "12" end,
						set = function(info, value) sRaidFrames.opt.Layouts[i].hpbar.height = tonumber(value); self:LayoutUpdate(i); end,
						order = 1,
					},
					width = {
						name = L["Width"],
						type = "input",
						desc = string.format(L["Modify the %s of this layout's %s bar"], L["width"], L["HP"]),
						get = function() return tostring(sRaidFrames.opt.Layouts[i].hpbar.width) or "80" end,
						set = function(info, value) sRaidFrames.opt.Layouts[i].hpbar.width = tonumber(value); self:LayoutUpdate(i); end,
						order = 2,
					},
					offsets = {
						name = L["Offsets"],
						type = "group",
						dialogInline = true,
						desc = string.format(L["Modify the %s of this layout's %s bar"], L["offsets"], L["HP"]),
						order = 3,
						args = {
							x = {
								name = "X",
								type = "input",
								desc = string.format(L["Modify the %s of this layout's %s bar"], "X "..L["offset"], L["HP"]),
								get = function() return tostring(sRaidFrames.opt.Layouts[i].hpbar.offsetx) or "0" end,
								set = function(info, value) sRaidFrames.opt.Layouts[i].hpbar.offsetx = tonumber(value); self:LayoutUpdate(i); end,
								order = 1,
							},
							y = {
								name = "Y",
								type = "input",
								desc = string.format(L["Modify the %s of this layout's %s bar"], "Y "..L["offset"], L["HP"]),
								get = function() return tostring(sRaidFrames.opt.Layouts[i].hpbar.offsety) or "0" end,
								set = function(info, value) sRaidFrames.opt.Layouts[i].hpbar.offsety = tonumber(value); self:LayoutUpdate(i); end,
								order = 2,
							},
						},
					},
				},
			},
			mpbar = {
				name = L["Mana bar"],
				type = "group",
				desc = string.format(L["Configure this layout's %s bar"], L["Mana"]),
				order = 3,
				args = {
					height = {
						name = L["Height"],
						type = "input",
						desc = string.format(L["Modify the %s of this layout's %s bar"], L["height"], L["Mana"]),
						get = function() return tostring(sRaidFrames.opt.Layouts[i].mpbar.height) or "4" end,
						set = function(info, value) sRaidFrames.opt.Layouts[i].mpbar.height = tonumber(value); self:LayoutUpdate(i); end,
					},
				},
			},
			buffsdebuffs = {
				name = L["Buffs & Debuffs"],
				type = "group",
				order = 4,
				desc = string.format(L["Configure this layout's %s"], L["Buffs & Debuffs"]),
				args = {
					anchor = {
						name = L["Anchor"],
						type = "select",
						desc = string.format(L["Configure the %s of this layout's %s"], L["anchor"], L["Buffs & Debuffs"]),
						get = function() return sRaidFrames.opt.Layouts[i].debuff.anchor or "TOPRIGHT" end,
						set = function(info, value) sRaidFrames.opt.Layouts[i].debuff.anchor = value; self:LayoutUpdate(i); end,
						values = {["TOPRIGHT"] = L["Top right"], ["TOPLEFT"] = L["Top left"], ["BOTTOMRIGHT"] = L["Bottom right"], ["BOTTOMLEFT"] = L["Bottom left"]},
						order = 1,
					},
					offsets = {
						name = L["Offsets"],
						type = "group",
						desc = string.format(L["Configure the %s of this layout's %s"], L["offsets"], L["Buffs & Debuffs"]),
						dialogInline = true,
						order = 2,
						args = {
							x = {
								name = "X",
								type = "input",
								desc = string.format(L["Configure the %s of this layout's %s"], "X "..L["offset"], L["Buffs & Debuffs"]),
								order = 1,
								get = function() return tostring(sRaidFrames.opt.Layouts[i].debuff.offsetx) end,
								set = function(info, value) sRaidFrames.opt.Layouts[i].debuff.offsetx = tonumber(value); self:LayoutUpdate(i); end,
							},
							y = {
								name = "Y",
								type = "input",
								desc = string.format(L["Configure the %s of this layout's %s"], "Y "..L["offset"], L["Buffs & Debuffs"]),
								order = 2,
								get = function() return tostring(sRaidFrames.opt.Layouts[i].debuff.offsety) end,
								set = function(info, value) sRaidFrames.opt.Layouts[i].debuff.offsety = tonumber(value); self:LayoutUpdate(i); end,
							},
						},
					},
					buffs = {
						name = L["Buffs"],
						type = "group",
						desc = string.format(L["Configure this layout's %s"], L["Buffs"]),
						args = {
							width = {
								name = L["Width"],
								type = "input",
								desc = string.format(L["Modify the %s of this layout's %s"], L["width"], L["Buffs"]),
								get = function() return tostring(sRaidFrames.opt.Layouts[i].buff.width) or "12" end,
								set = function(info, value) sRaidFrames.opt.Layouts[i].buff.width = tonumber(value); self:LayoutUpdate(i); end,
							},
							height = {
								name = L["Height"],
								type = "input",
								desc = string.format(L["Modify the %s of this layout's %s"], L["height"], L["Buffs"]),
								get = function() return tostring(sRaidFrames.opt.Layouts[i].buff.height) or "12" end,
								set = function(info, value) sRaidFrames.opt.Layouts[i].buff.height = tonumber(value); self:LayoutUpdate(i); end,
							},
						},
					},
					debuffs = {
						name = L["Debuffs"],
						type = "group",
						desc = string.format(L["Configure this layout's %s"], L["Debuffs"]),
						args = {
							width = {
								name = L["Width"],
								type = "input",
								desc = string.format(L["Modify the %s of this layout's %s"], L["width"], L["Debuffs"]),
								get = function() return tostring(sRaidFrames.opt.Layouts[i].debuff.width) or "16" end,
								set = function(info, value) sRaidFrames.opt.Layouts[i].debuff.width = value; self:LayoutUpdate(i); end,
							},
							height = {
								name = L["Height"],
								type = "input",
								desc = string.format(L["Modify the %s of this layout's %s"], L["height"], L["Debuffs"]),
								get = function() return tostring(sRaidFrames.opt.Layouts[i].debuff.height) or "16" end,
								set = function(info, value) sRaidFrames.opt.Layouts[i].debuff.height = value; self:LayoutUpdate(i); end,
							},
						},
					},
				},
			},
			font = {
				name = L["Status Text"],
				type = "group",
				desc = string.format(L["Modify this layout's %s"], L["Status text"]),
				order = 5,
				args = {
					font = {
						name = L["Font"],
						type = "select",
						desc = string.format(L["Modify the %s of this layout's %s"], L["Font"], L["Status text"]),
						get = function() return sRaidFrames.opt.Layouts[i].font.font end,
						set = function(info, value) sRaidFrames.opt.Layouts[i].font.font = value; self:LayoutUpdate(i); end,
						values = Media:HashTable("font"),
						dialogControl = "LSM30_Font",
						order = 1,
					},
					fontscale = {
						name = L["Font scale"],
						type = "range",
						desc = string.format(L["Modify the %s of this layout's %s"], L["Font scale"], L["Status text"]),
						min = 1,
						max = 50,
						step = 1,
						get = function() return sRaidFrames.opt.Layouts[i].font.scale end,
						set = function(info, value) sRaidFrames.opt.Layouts[i].font.scale = value; self:LayoutUpdate(i); end,
						order = 2,
					},
					fontoutline = {
						name = L["Font outline"],
						type = "select",
						desc = string.format(L["Modify the %s of this layout's %s"], L["Font outline"], L["Status text"]),
						values = {["OUTLINE"] = L["Outline"], ["THICKOUTLINE"] = L["Thick outline"], ["MONOCHROME"] = L["Monochrome"], [""] = L["None"]},
						get = function() return sRaidFrames.opt.Layouts[i].font.outline end,
						set = function(info, value) sRaidFrames.opt.Layouts[i].font.outline = value; self:LayoutUpdate(i); end,
						order = 3,
					},
				},
			},
			indicator = {
				name = L["Indicators"],
				type = "group",
				desc = string.format(L["Modify this layout's %s"], L["Indicators"]),
				order = 6,
				args = {
					scale = {
						name = L["Scale"],
						type = "range",
						desc = string.format(L["Modify the %s of this layout's %s"], L["scale"], L["Indicators"]),
						min = 0.1,
						max = 2.0,
						get = function() return sRaidFrames.opt.Layouts[i].indicator.scale end,
						set = function(info, value) sRaidFrames.opt.Layouts[i].indicator.scale = value; self:LayoutUpdate(i); end,
						isPercent = true,
					},
				},
			},
			border = {
				name = L["Border"],
				type = "toggle",
				desc = L["Select whether this layout uses a border"],
				get = function() return sRaidFrames.opt.Layouts[i].border or false end,
				set = function(info, value) sRaidFrames.opt.Layouts[i].border = value; self:LayoutUpdate(i); end,
				order = 1,
			},
			bordertexture = {
				name = L["Border texture"],
				type = "select",
				desc = string.format(L["Configure this layout's %s"], L["Border texture"]),
				get = function()
					return sRaidFrames.opt.Layouts[i].bordertexture
				end,
				set = function(info, value)
					sRaidFrames.opt.Layouts[i].bordertexture = value
					if not sRaidFrames.opt.Layout == sRaidFrames.opt.Layouts[i].intname then return end
					local tex = Media:Fetch("border", sRaidFrames.opt.Layouts[i].bordertexture)
					for _, frame in pairs(sRaidFrames.frames) do
						local backdrop = frame:GetBackdrop()
						backdrop.edgeFile = tex
						frame:SetBackdrop(backdrop)
					end
					sRaidFrames:UpdateAllUnits()
				end,
				values = AceGUIWidgetLSMlists.border,
				order = 2,
				dialogControl = 'LSM30_Border',
			},
		},
	}
end

function sRaidFrames:GetLayoutDefaults(name)
	local a= {
		bordertexture = sRaidFrames.opt.BorderTexture,
		border = false,
		name = name,
		intname = string.gsub(string.lower(name), " ", ""),
		debuff = {
			width = 16,
			height = 16,
			anchor = "TOPRIGHT",
			offsetx = 0,
			offsety = 0,
		},
		buff = {
			width = 12,
			height = 12,
		},
		title = {
			width = 80,
			height = 16,
			offsetx = 0,
			offsety = 0,
		},
		mpbar = {
			height = 4,
		},
		hpbar = {
			height = 12,
			width = 80,
			offsetx = 0,
			offsety = 0,
		},
		font = {
			font = "Friz Quadrata TT",
			scale = 10.000000140577,
			outline = "",
		},
		indicator = {
			scale = 1,
		},
	}
	return a
end
local function TableCompare(stable, default)
	for x, k in pairs(default) do
		if type(k)=="table" then
			if not stable[x] then stable[x] = {}; end
			stable[x] = TableCompare(stable[x], default[x]);
		else
			if not stable[x] then stable[x] = default[x] end
		end
	end
	return stable
end

function sRaidFrames:VerifyLayout(layout)
	if not layout.name then return end
	local default = sRaidFrames:GetLayoutDefaults(layout.name)
	return TableCompare(layout, default)
end

function sRaidFrames:CreateLayout(name)
	local i = #sRaidFrames.opt.Layouts+1;
	sRaidFrames.opt.Layouts[i] = self:GetLayoutDefaults(name);
	local data = sRaidFrames.opt.Layouts[i];
	sRaidFrames:RegisterLayout(data.intname, {
		Name = data.name,
		HasBorders = data.border,
		StyleUnitFrame = function(f)
			sRaidFrames:SetWHP(f, data.border and data.title.width+10 or data.title.width, ((data.debuff.anchor=="BOTTOMLEFT" or data.debuff.anchor=="BOTTOMRIGHT") and (data.mpbar.height+data.hpbar.height+math.max(data.debuff.height, data.buff.height)+data.title.height) or (data.mpbar.height+data.hpbar.height+data.title.height))+(data.border and 10 or 0));
			sRaidFrames:SetWHP(f.title, data.title.width, data.title.height, "TOPLEFT", f, "TOPLEFT", data.border and data.title.offsetx+5 or data.title.offsetx, data.border and data.title.offsety-4 or data.title.offsety)
			sRaidFrames:SetWHP(f.aura1, data.debuff.width, data.debuff.height, data.debuff.anchor, f, data.debuff.anchor, ((data.border and (data.debuff.anchor=="TOPRIGHT" or data.debuff.anchor=="BOTTOMRIGHT")) and -4 or data.border and 4 or 0)+(data.debuff.offsetx or 0), ((data.border and (data.debuff.anchor=="TOPRIGHT" or data.debuff.anchor=="TOPLEFT")) and -4 or data.border and 4 or 0)+(data.debuff.offsety or 0))
			sRaidFrames:SetWHP(f.aura2, data.debuff.width, data.debuff.height, (data.debuff.anchor=="TOPLEFT" or data.debuff.anchor=="BOTTOMLEFT") and "LEFT" or "RIGHT", f.aura1, (data.debuff.anchor=="TOPLEFT" or data.debuff.anchor=="BOTTOMLEFT") and "RIGHT" or "LEFT", 0, 0)
			if sRaidFrames.opt.BuffType == "both" then
				local parent
				if f.aura2:IsShown() then
					parent = f.aura2
				elseif f.aura1:IsShown() then
					parent = f.aura1
				end
				if parent then
					sRaidFrames:SetWHP(f.buff1, data.buff.width, data.buff.height, data.debuff.anchor, parent, data.debuff.anchor=="TOPRIGHT" and "TOPLEFT" or data.debuff.anchor=="TOPLEFT" and "TOPRIGHT" or data.debuff.anchor=="BOTTOMRIGHT" and "BOTTOMLEFT" or data.debuff.anchor=="BOTTOMLEFT" and "BOTTOMRIGHT", 0, 0)
				else
					sRaidFrames:SetWHP(f.buff1, data.buff.width, data.buff.height, data.debuff.anchor, f, data.debuff.anchor, ((data.border and (data.debuff.anchor=="TOPRIGHT" or data.debuff.anchor=="BOTTOMRIGHT")) and -4 or data.border and 4 or 0)+(data.debuff.offsetx or 0), ((data.border and (data.debuff.anchor=="TOPRIGHT" or data.debuff.anchor=="BOTTOMRIGHT")) and -4 or data.border and 4 or 0)+(data.debuff.offsety or 0))
				end
			else
				sRaidFrames:SetWHP(f.buff1, data.buff.width, data.buff.height, data.debuff.anchor, f, data.debuff.anchor, ((data.border and (data.debuff.anchor=="TOPRIGHT" or data.debuff.anchor=="BOTTOMRIGHT")) and -4 or data.border and 4 or 0)+(data.debuff.offsetx or 0), ((data.border and (data.debuff.anchor=="TOPRIGHT" or data.debuff.anchor=="TOPLEFT")) and -4 or data.border and 4 or 0)+(data.debuff.offsety or 0))
			end
			sRaidFrames:SetWHP(f.buff2, data.buff.width, data.buff.height, (data.debuff.anchor=="TOPLEFT" or data.debuff.anchor=="BOTTOMLEFT") and "LEFT" or "RIGHT", f.buff1, (data.debuff.anchor=="TOPLEFT" or data.debuff.anchor=="BOTTOMLEFT") and "RIGHT" or "LEFT", 0, 0)
			sRaidFrames:SetWHP(f.buff3, data.buff.width, data.buff.height, (data.debuff.anchor=="TOPLEFT" or data.debuff.anchor=="BOTTOMLEFT") and "LEFT" or "RIGHT", f.buff2, (data.debuff.anchor=="TOPLEFT" or data.debuff.anchor=="BOTTOMLEFT") and "RIGHT" or "LEFT", 0, 0)
			sRaidFrames:SetWHP(f.buff4, data.buff.width, data.buff.height, (data.debuff.anchor=="TOPLEFT" or data.debuff.anchor=="BOTTOMLEFT") and "LEFT" or "RIGHT", f.buff3, (data.debuff.anchor=="TOPLEFT" or data.debuff.anchor=="BOTTOMLEFT") and "RIGHT" or "LEFT", 0, 0)
			sRaidFrames:SetWHP(f.hpbar, data.hpbar.width, data.hpbar.height, "TOPLEFT", f.title, "BOTTOMLEFT", data.hpbar.offsetx, data.hpbar.offsety)
			sRaidFrames:SetWHP(f.mpbar, data.hpbar.width, data.mpbar.height, "TOPLEFT", f.hpbar, "BOTTOMLEFT", 0, 0)
			sRaidFrames:SetWHP(f.hpbar.text, f.hpbar:GetWidth(), f.hpbar:GetHeight(), "CENTER", f.hpbar, "CENTER", 0, 0)
			sRaidFrames:SetWHP(f.statustext, f.mpbar:GetWidth(), f.mpbar:GetHeight(), "CENTER", f.mpbar, "CENTER", 0, 0)
			f:SetBackdrop(data.border and { bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
									tile = true,
									tileSize = 16,
									edgeFile = Media:Fetch("border", data.bordertexture),
									edgeSize = 16,
									insets = { left = 5, right = 5, top = 5, bottom = 5 }
								} or { bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
										tile = true, tileSize = 16,
										insets = { left = 0, right = 0, top = 0, bottom = 0 }
									})
			f:SetBackdropColor(sRaidFrames.opt.BackgroundColor.r, sRaidFrames.opt.BackgroundColor.g, sRaidFrames.opt.BackgroundColor.b, sRaidFrames.opt.BackgroundColor.a)
			f.hpbar.text:SetTextColor(sRaidFrames.opt.HealthTextColor.r, sRaidFrames.opt.HealthTextColor.g, sRaidFrames.opt.HealthTextColor.b, sRaidFrames.opt.HealthTextColor.a)
			f.statustext:SetFont(Media:Fetch("font", data.font.font), data.font.scale, data.font.outline)
			if IsAddOnLoaded("sRaidFramesIndicators") then
				DisableAddOn("sRaidFramesIndicators");
				ReloadUI();
			end
			if not f.indicator_tl then
				sRaidFrames:CreateIndicators(f)
			end
			f.indicator_tl:SetFrameLevel(6)
			f.indicator_tr:SetFrameLevel(6)
			f.indicator_bl:SetFrameLevel(6)
			f.indicator_br:SetFrameLevel(6)
			sRaidFrames:SetWHP(f.indicator_tl, 8*(data.indicator.scale or 1), 8*(data.indicator.scale or 1), "TOPLEFT", f, "TOPLEFT", -1*(data.indicator.scale or 1), 1*(data.indicator.scale or 1))
			sRaidFrames:SetWHP(f.indicator_tr, 8*(data.indicator.scale or 1), 8*(data.indicator.scale or 1), "TOPRIGHT", f, "TOPRIGHT", 1*(data.indicator.scale or 1), 1*(data.indicator.scale or 1));
			sRaidFrames:SetWHP(f.indicator_bl, 8*(data.indicator.scale or 1), 8*(data.indicator.scale or 1), "BOTTOMLEFT", f, "BOTTOMLEFT", -1*(data.indicator.scale or 1), -1*(data.indicator.scale or 1))
			sRaidFrames:SetWHP(f.indicator_br, 8*(data.indicator.scale or 1), 8*(data.indicator.scale or 1), "BOTTOMRIGHT", f, "BOTTOMRIGHT", 1*(data.indicator.scale or 1), -1*(data.indicator.scale or 1))
		end,
		unitframeWidth = data.border and data.title.width+10 or data.title.width,
		unitframeHeight =  ((data.debuff.anchor=="BOTTOMLEFT" or data.debuff.anchor=="BOTTOMRIGHT") and (data.mpbar.height+data.hpbar.height+math.max(data.debuff.height, data.buff.height)+data.title.height) or (data.mpbar.height+data.hpbar.height+data.title.height))+(data.border and 10 or 0),
		headerWidth = data.border and data.title.width+10 or data.title.width,
		headerHeight = 15, -- needs more testing
		debuffCount = 2,
		buffCount = 4,
	})
	self:CreateLayoutOptions(data, i);
end

function sRaidFrames:LayoutUpdate(i)
	local data = sRaidFrames.opt.Layouts[i]
	self:CreateLayoutOptions(data, i)
	self.options.args.behaviour.args.layout.values = {}
	for _name, _data in pairs(self.Layouts) do
		self.options.args.behaviour.args.layout.values[_name] = _data.Name
	end
	if data.intname == sRaidFrames.opt.Layout then self:ApplyLayout() end
end

function sRaidFrames:GetCurrentCustomLayoutID()
	local name = self:GetLayout()["Name"];
	for i, k in pairs(self.opt.Layouts) do
		if k["name"] == name then
			return i
		end
	end
end

function sRaidFrames:RegisterCustomLayouts()
	for i=1, #sRaidFrames.opt.Layouts do
		sRaidFrames.opt.Layouts[i] = self:VerifyLayout(sRaidFrames.opt.Layouts[i])
		local data = sRaidFrames.opt.Layouts[i]
		if data then
			sRaidFrames:RegisterLayout(data.intname, {
				Name = data.name,
				HasBorders = data.border,
				StyleUnitFrame = function(f)
					sRaidFrames:SetWHP(f, data.border and data.title.width+10 or data.title.width,  ((data.debuff.anchor=="BOTTOMLEFT" or data.debuff.anchor=="BOTTOMRIGHT") and (data.mpbar.height+data.hpbar.height+math.max(data.debuff.height, data.buff.height)+data.title.height) or (data.mpbar.height+data.hpbar.height+data.title.height))+(data.border and 10 or 0));
					sRaidFrames:SetWHP(f.title, data.title.width, data.title.height, "TOPLEFT", f, "TOPLEFT", data.border and data.title.offsetx+5 or data.title.offsetx, data.border and data.title.offsety-4 or data.title.offsety)
					sRaidFrames:SetWHP(f.aura1, data.debuff.width, data.debuff.height, data.debuff.anchor, f, data.debuff.anchor, ((data.border and (data.debuff.anchor=="TOPRIGHT" or data.debuff.anchor=="BOTTOMRIGHT")) and -4 or data.border and 4 or 0)+(data.debuff.offsetx or 0), ((data.border and (data.debuff.anchor=="TOPRIGHT" or data.debuff.anchor=="TOPLEFT")) and -4 or data.border and 4 or 0)+(data.debuff.offsety or 0))
					sRaidFrames:SetWHP(f.aura2, data.debuff.width, data.debuff.height, (data.debuff.anchor=="TOPLEFT" or data.debuff.anchor=="BOTTOMLEFT") and "LEFT" or "RIGHT", f.aura1, (data.debuff.anchor=="TOPLEFT" or data.debuff.anchor=="BOTTOMLEFT") and "RIGHT" or "LEFT", 0, 0)
					if sRaidFrames.opt.BuffType == "both" then
						local parent
						if f.aura2:IsShown() then
							parent = f.aura2
						elseif f.aura1:IsShown() then
							parent = f.aura1
						end
						if parent then
							sRaidFrames:SetWHP(f.buff1, data.buff.width, data.buff.height, data.debuff.anchor, parent, data.debuff.anchor=="TOPRIGHT" and "TOPLEFT" or data.debuff.anchor=="TOPLEFT" and "TOPRIGHT" or data.debuff.anchor=="BOTTOMRIGHT" and "BOTTOMLEFT" or data.debuff.anchor=="BOTTOMLEFT" and "BOTTOMRIGHT", 0, 0)
						else
							sRaidFrames:SetWHP(f.buff1, data.buff.width, data.buff.height, data.debuff.anchor, f, data.debuff.anchor, ((data.border and (data.debuff.anchor=="TOPRIGHT" or data.debuff.anchor=="BOTTOMRIGHT")) and -4 or data.border and 4 or 0)+(data.debuff.offsetx or 0), ((data.border and (data.debuff.anchor=="TOPRIGHT" or data.debuff.anchor=="BOTTOMRIGHT")) and -4 or data.border and 4 or 0)+(data.debuff.offsety or 0))
						end
					else
						sRaidFrames:SetWHP(f.buff1, data.buff.width, data.buff.height, data.debuff.anchor, f, data.debuff.anchor, ((data.border and (data.debuff.anchor=="TOPRIGHT" or data.debuff.anchor=="BOTTOMRIGHT")) and -4 or data.border and 4 or 0)+(data.debuff.offsetx or 0), ((data.border and (data.debuff.anchor=="TOPRIGHT" or data.debuff.anchor=="TOPLEFT")) and -4 or data.border and 4 or 0)+(data.debuff.offsety or 0))
					end
					sRaidFrames:SetWHP(f.buff2, data.buff.width, data.buff.height, (data.debuff.anchor=="TOPLEFT" or data.debuff.anchor=="BOTTOMLEFT") and "LEFT" or "RIGHT", f.buff1, (data.debuff.anchor=="TOPLEFT" or data.debuff.anchor=="BOTTOMLEFT") and "RIGHT" or "LEFT", 0, 0)
					sRaidFrames:SetWHP(f.buff3, data.buff.width, data.buff.height, (data.debuff.anchor=="TOPLEFT" or data.debuff.anchor=="BOTTOMLEFT") and "LEFT" or "RIGHT", f.buff2, (data.debuff.anchor=="TOPLEFT" or data.debuff.anchor=="BOTTOMLEFT") and "RIGHT" or "LEFT", 0, 0)
					sRaidFrames:SetWHP(f.buff4, data.buff.width, data.buff.height, (data.debuff.anchor=="TOPLEFT" or data.debuff.anchor=="BOTTOMLEFT") and "LEFT" or "RIGHT", f.buff3, (data.debuff.anchor=="TOPLEFT" or data.debuff.anchor=="BOTTOMLEFT") and "RIGHT" or "LEFT", 0, 0)
					sRaidFrames:SetWHP(f.hpbar, data.hpbar.width, data.hpbar.height, "TOPLEFT", f.title, "BOTTOMLEFT", data.hpbar.offsetx, data.hpbar.offsety)
					sRaidFrames:SetWHP(f.mpbar, data.hpbar.width, data.mpbar.height, "TOPLEFT", f.hpbar, "BOTTOMLEFT", 0, 0)
					sRaidFrames:SetWHP(f.hpbar.text, f.hpbar:GetWidth(), f.hpbar:GetHeight(), "CENTER", f.hpbar, "CENTER", 0, 0)
					sRaidFrames:SetWHP(f.statustext, f.mpbar:GetWidth(), f.mpbar:GetHeight(), "CENTER", f.mpbar, "CENTER", 0, 0)
					f:SetBackdrop(data.border and { bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
											tile = true,
											tileSize = 16,
											edgeFile = Media:Fetch("border", data.bordertexture),
											edgeSize = 16,
											insets = { left = 5, right = 5, top = 5, bottom = 5 }
										} or { bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
												tile = true, tileSize = 16,
												insets = { left = 0, right = 0, top = 0, bottom = 0 }
											})
					f:SetBackdropColor(sRaidFrames.opt.BackgroundColor.r, sRaidFrames.opt.BackgroundColor.g, sRaidFrames.opt.BackgroundColor.b, sRaidFrames.opt.BackgroundColor.a)
					f.hpbar.text:SetTextColor(sRaidFrames.opt.HealthTextColor.r, sRaidFrames.opt.HealthTextColor.g, sRaidFrames.opt.HealthTextColor.b, sRaidFrames.opt.HealthTextColor.a)
					f.statustext:SetFont(Media:Fetch("font", data.font.font), data.font.scale, data.font.outline)
					if IsAddOnLoaded("sRaidFramesIndicators") and not f.indicator_init then
						DisableAddOn("sRaidFramesIndicators")
						ReloadUI()
					end
					if not f.indicator_tl then
						sRaidFrames:CreateIndicators(f)
					end
					f.indicator_tl:SetFrameLevel(6)
					f.indicator_tr:SetFrameLevel(6)
					f.indicator_bl:SetFrameLevel(6)
					f.indicator_br:SetFrameLevel(6)
					sRaidFrames:SetWHP(f.indicator_tl, 8*(data.indicator.scale or 1), 8*(data.indicator.scale or 1), "TOPLEFT", f, "TOPLEFT", -1*(data.indicator.scale or 1), 1*(data.indicator.scale or 1))
					sRaidFrames:SetWHP(f.indicator_tr, 8*(data.indicator.scale or 1), 8*(data.indicator.scale or 1), "TOPRIGHT", f, "TOPRIGHT", 1*(data.indicator.scale or 1), 1*(data.indicator.scale or 1));
					sRaidFrames:SetWHP(f.indicator_bl, 8*(data.indicator.scale or 1), 8*(data.indicator.scale or 1), "BOTTOMLEFT", f, "BOTTOMLEFT", -1*(data.indicator.scale or 1), -1*(data.indicator.scale or 1))
					sRaidFrames:SetWHP(f.indicator_br, 8*(data.indicator.scale or 1), 8*(data.indicator.scale or 1), "BOTTOMRIGHT", f, "BOTTOMRIGHT", 1*(data.indicator.scale or 1), -1*(data.indicator.scale or 1))
				end,
				unitframeWidth = data.border and data.title.width+10 or data.title.width,
				unitframeHeight =  ((data.debuff.anchor=="BOTTOMLEFT" or data.debuff.anchor=="BOTTOMRIGHT") and (data.mpbar.height+data.hpbar.height+math.max(data.debuff.height, data.buff.height)+data.title.height) or (data.mpbar.height+data.hpbar.height+data.title.height))+(data.border and 10 or 0),
				headerWidth = data.border and data.title.width+10 or data.title.width,
				headerHeight = 15, -- needs more testing
				debuffCount = 2,
				buffCount = 4,
			})
			self:CreateLayoutOptions(data, i)
		end
	end
end

--Initialize options
if sRaidFrames.opt.advanced == nil then
	sRaidFrames.opt.advanced = false
end

if sRaidFrames.opt.PredictedHealth == nil then
	sRaidFrames.opt.PredictedHealth = true
end

if not sRaidFrames.opt.Layouts then
	sRaidFrames.opt.Layouts = {}
end

sRaidFrames.options.args.expert = {
	type = "execute",
	order = 6,
	name = function() return sRaidFrames.opt.advanced and L["Switch to default configuration"] or L["Switch to advanced configuration"] end,
	func = function() sRaidFrames.opt.advanced = not sRaidFrames.opt.advanced for i in pairs(sRaidFrames.opt.StatusMaps) do sRaidFrames.options.args.advanced.args[i].args.color.hasAlpha = sRaidFrames.opt.advanced end end,
	desc = function() return sRaidFrames.opt.advanced and L["In default mode, only basic configuration options are available"] or L["In advanced mode, additional customization options are available"] end,
	width = "double",
}
		
sRaidFrames.options.args.behaviour.args.additional = {
	name = L["Additional Options"],
	type = "group",
	desc = L["Choose additional options"],
	dialogInline = true,
	args = {
		predictedhealth = {
			name = L["Predicted Health"],
			type = "toggle",
			desc = L["Choose whether you want to enable PredictedHealth. This updates Health Bars faster, but increases memory usage."],
			get = function() return sRaidFrames.opt.PredictedHealth end,
			set = function(info, value) sRaidFrames.opt.PredictedHealth = value; if value then sRaidFrames.predictedhealthframe:SetScript("OnUpdate", sRaidFrames.PredictedHealthOnUpdate); else sRaidFrames.predictedhealthframe:SetScript("OnUpdate", nil); end end,
			order = 1,
		},
	},
	order = 60,
	hidden = function() return not sRaidFrames.opt.advanced end,
}

sRaidFrames.options.args.layouts = {
	name = L["Layouts"],
	type = "group",
	desc = L["Define custom layouts"],
	hidden = function() return not sRaidFrames.opt.advanced end,
	args = {
		create = {
			name = L["Create layout"],
			type = "input",
			desc = L["Create a new layout"],
			get = function() if not sRaidFrames.layoutoptions then sRaidFrames.layoutoptions = {}; end return sRaidFrames.layoutoptions.createname or false end,
			set = function(info, value) sRaidFrames.layoutoptions.createname = value end,
			order = 1,
		},
		createbutton = {
			name = L["Create layout"],
			type = "execute",
			desc = L["Create the layout"],
			func = function() sRaidFrames:CreateLayout(sRaidFrames.layoutoptions.createname) sRaidFrames.layoutoptions.createname = nil; end,
			disabled = function() return not sRaidFrames.layoutoptions.createname end,
			order = 2,
		},
		remove = {
			name = L["Remove layout"],
			type = "select",
			desc = L["Remove a custom layout"],
			get = function() if not sRaidFrames.layoutoptions then sRaidFrames.layoutoptions = {} end return sRaidFrames.layoutoptions.deletename or false end,
			set = function(info, value) sRaidFrames.layoutoptions.deletename = value end,
			values = function() local a = {}; for i, k in pairs(sRaidFrames.opt.Layouts) do if sRaidFrames.opt.Layout~=k.intname then a[i] = k.name; end end return a end,
			order = 3,
		},
		removebutton = {
			name = L["Remove layout"],
			type = "execute",
			desc = L["Remove this custom layout. WARNING: THIS CAN NOT BE REVERTED!"],
			func = function() local data = sRaidFrames.opt.Layouts[sRaidFrames.layoutoptions.deletename]; sRaidFrames.Layouts[data.intname] = {}; sRaidFrames.opt.Layouts[sRaidFrames.layoutoptions.deletename] = nil; sRaidFrames.options.args.layouts.args[data.intname] = nil end,
			disabled = function() return not sRaidFrames.layoutoptions.deletename end,
			order = 4,
		},
	},
	disabled = InCombatLockdown,
	order = 275,
}

sRaidFrames.options.args.buffsdebuffs.args.debuffs.args.blacklist = {
	name = L["Buff Display settings"],
	type = "group",
	desc = L["Set when to display certain buffs"],
	dialogInline = true,
	args = {
		name = {
			name = L["Buff"],
			type = "input",
			desc = L["Select a buff to configure"],
			get = function() return sRaidFrames.current_buff or false end,
			set = function(info, value)
				if tonumber(value) then
					value = GetSpellInfo(value)
				end
				if value and value ~= "" then
					sRaidFrames.current_buff = value
				end
			end,
			dialogControl = "Aura_EditBox",
			usage = L["<name of buff>"],
			order = 1,
		},
		suggest = {
			name = L["Suggestions"],
			type = "select",
			desc = L["Already configured buffs"],
			get = function() return sRaidFrames.current_buff or false end,
			set = function(info, value) sRaidFrames.current_buff = value; end,
			values = function() local t for buff in pairs(sRaidFrames.opt.BuffDisplayOptions) do if buff then if not t then t = {}; end t[buff] = buff; end end if sRaidFrames.current_buff and (not t or not t[string.lower(sRaidFrames.current_buff)]) then if not t then t = {}; end t[string.lower(sRaidFrames.current_buff)] = string.lower(sRaidFrames.current_buff); end return t or {} end,
			order = 2,
		},
		--values = {[0] = L["Never"], [1] = L["In Combat"], [2] = L["Out of combat"], [3] = L["Always"]},
		set = {
			name = L["Adjust Display Setting"],
			type = "group",
			desc = L["Select when to display this"],
			dialogInline = true,
			args = {
				never = {
					name = L["Never"],
					type = "toggle",
					desc = L["Never show this"],
					get = function() if sRaidFrames.opt.BuffDisplayOptions[string.lower(sRaidFrames.current_buff)] == 0 then return true else return false end end,
					set = function(info, value) if value then sRaidFrames.opt.BuffDisplayOptions[string.lower(sRaidFrames.current_buff)] = 0 for unit in pairs(sRaidFrames:GetAllUnits()) do sRaidFrames:UpdateAuras(unit) end end end,
					order = 1,
				},
				incombat = {
					name = L["In Combat"],
					type = "toggle",
					desc = L["Show this in combat"],
					get = function() if sRaidFrames.opt.BuffDisplayOptions[string.lower(sRaidFrames.current_buff)] == 1 then return true else return false end end,
					set = function(info, value) if value then sRaidFrames.opt.BuffDisplayOptions[string.lower(sRaidFrames.current_buff)] = 1 for unit in pairs(sRaidFrames:GetAllUnits()) do sRaidFrames:UpdateAuras(unit) end end end,
					order = 2,
				},
				outofcombat = {
					name = L["Out of combat"],
					type = "toggle",
					desc = L["Show this out of combat"],
					get = function() if sRaidFrames.opt.BuffDisplayOptions[string.lower(sRaidFrames.current_buff)] == 2 then return true else return false end end,
					set = function(info, value) if value then sRaidFrames.opt.BuffDisplayOptions[string.lower(sRaidFrames.current_buff)] = 2 for unit in pairs(sRaidFrames:GetAllUnits()) do sRaidFrames:UpdateAuras(unit) end end end,
					order = 3,
				},
				always = {
					name = L["Always"],
					type = "toggle",
					desc = L["Always show this"],
					get = function() if sRaidFrames.opt.BuffDisplayOptions[string.lower(sRaidFrames.current_buff)] == 3 or not sRaidFrames.opt.BuffDisplayOptions[string.lower(sRaidFrames.current_buff)] then return true else return false end end,
					set = function(info, value) if value then sRaidFrames.opt.BuffDisplayOptions[string.lower(sRaidFrames.current_buff)] = 3 for unit in pairs(sRaidFrames:GetAllUnits()) do sRaidFrames:UpdateAuras(unit) end end end,
					order = 4,
				},
			},
			order = 3,
			hidden = function() return not sRaidFrames.current_buff and true or false end,
		},
	},
	order = 1,
	hidden = function() return not sRaidFrames.opt.advanced end,
}
sRaidFrames.options.args.buffsdebuffs.args.buffs.args.oldblacklist.hidden = function() return sRaidFrames.opt.advanced end

function sRaidFrames:chatUpdateStatusElements()
	local i = 1
	for key, data in pairs(sRaidFrames.opt.StatusMaps) do
		local name = key
		if key:find("^Buff_") then
			local id = key:match("^Buff_(%d+)")
			name = GetSpellInfo(id)
		end
		if not sRaidFrames.options.args.advanced.args[key] then
			sRaidFrames.options.args.advanced.args[key] = {
				type = 'group',
				name = name,
				desc = ("Change settings for the status %s"):format(name),
				args = {
					priority = {
						name = L["Priority"],
						desc = L["Set the priority"],
						type = "range",
						min = 1,
						max = 200,
						step = 1,
						set = function(info, value) sRaidFrames.opt.StatusMaps[key].priority = value end,
						get = function() return sRaidFrames.opt.StatusMaps[key].priority end,
					},
					enabled = {
						name = L["Enabled"],
						desc = L["Enable the status indicator"],
						type = "toggle",
						set = function(info, value) sRaidFrames.opt.StatusMaps[key].enabled = value end,
						get = function() return sRaidFrames.opt.StatusMaps[key].enabled ~= false end,
					},
					elements = {
						name = L["GUI Elements"],
						desc = L["Set which elements this status will use"],
						type = "multiselect",
						values = sRaidFrames.validateStatusElements,
						set = function(info, element, value) sRaidFrames.opt.StatusMaps[key].elements[element] = value end,
						get = function(info, element) return sRaidFrames.opt.StatusMaps[key].elements[element] end,
					},
					color = {
						name = L["Color"],
						desc = L["Set which color this status will use"],
						type = "color",
						set = function(info, r, g, b, a)
							sRaidFrames.opt.StatusMaps[key].color = {r = r, g = g, b = b, a = a}
						 end,
						get = function() 
							local color = sRaidFrames.opt.StatusMaps[key].color
							return color.r, color.g, color.b, color.a 
						end,
						hasAlpha = sRaidFrames.opt.advanced,
					},
					text = {
						name = L["Text"],
						desc = L["Set which text this status will show"],
						type = "input",
						usage = "<name>",
						set = function(info, value) sRaidFrames.opt.StatusMaps[key].text = value end,
						get = function() return sRaidFrames.opt.StatusMaps[key].text end,
					},
				},
				order = 200-sRaidFrames.opt.StatusMaps[key].priority,
			}
		end
	end
end

self:RegisterCustomLayouts()
self.predictedhealthframe = CreateFrame("Frame",nil,UIParent)
local pred = self.predictedhealthframe;
pred:Hide()
if self.opt.PredictedHealth then
	pred:SetScript("OnUpdate", self.PredictedHealthOnUpdate)
end
sRaidFrames.OriginalUpdateRoster = sRaidFrames.UpdateRoster
function sRaidFrames:UpdateRoster()
	sRaidFrames:OriginalUpdateRoster()
	self.predictedhealthframe[GetNumRaidMembers() > 0 and "Show" or "Hide"](self.predictedhealthframe)
end
sRaidFrames.OriginalUpdateAuras = sRaidFrames.UpdateAuras
function sRaidFrames:UpdateAuras(unit)
	sRaidFrames:OriginalUpdateAuras(unit)
	if not sRaidFrames.opt.BuffType == "both" then return end
	if not sRaidFrames:IsTracking(unit) then return end
	for _, f in pairs(sRaidFrames:FindUnitFrames(unit)) do
		local unit = sRaidFrames:GetVehicleUnit(unit)
		local data
		local id = sRaidFrames:GetCurrentCustomLayoutID()
		if id then
			data = self.opt.Layouts[id]
		end
		
		if data then
			local parent
			if f.aura2:IsShown() then
				parent = f.aura2;
			elseif f.aura1:IsShown() then
				parent = f.aura1;
			end
			local buffFrame = f.buff1
			if not parent then
				self:SetWHP(buffFrame, buffFrame:GetWidth(), buffFrame:GetHeight(), data.debuff.anchor, f, data.debuff.anchor, (data.border and (data.debuff.anchor=="TOPRIGHT" or data.debuff.anchor=="TOPLEFT")) and -4 or data.border and 4 or 0, ((data.border and (data.debuff.anchor=="TOPRIGHT" or data.debuff.anchor=="TOPLEFT")) and -4 or data.border and 4 or 0)+data.debuff.offsety);
			else
				local relativeto = (data.debuff.anchor == "BOTTOMLEFT" and "BOTTOMRIGHT" or data.debuff.anchor == "TOPLEFT" and "TOPRIGHT" or data.debuff.anchor == "BOTTOMRIGHT" and "BOTTOMLEFT" or data.debuff.anchor == "TOPRIGHT" and "TOPLEFT")
				self:SetWHP(buffFrame, buffFrame:GetWidth(), buffFrame:GetHeight(), data.debuff.anchor, parent, relativeto, 0, 0);
			end
		end
	end
end

sRaidFrames:RegisterStatusElement("indicator-tl", "Indicator (Top Left)",
	function(self, frame, status)
		if not frame.indicator_tl then return end
		if status == nil then
			frame.indicator_tl:SetBackdropColor(0,0,0,0)
		else
			frame.indicator_tl:SetBackdropColor(status.color.r, status.color.g, status.color.b, status.color.a or 1)
		end
	end)

sRaidFrames:RegisterStatusElement("indicator-tr", "Indicator (Top Right)",
	function(self, frame, status)
		if not frame.indicator_tr then return end
		if status == nil then
			frame.indicator_tr:SetBackdropColor(0,0,0,0)
		else
			frame.indicator_tr:SetBackdropColor(status.color.r, status.color.g, status.color.b, status.color.a or 1)
		end
	end)

sRaidFrames:RegisterStatusElement("indicator-bl", "Indicator (Bottom Left)",
	function(self, frame, status)
		if not frame.indicator_bl then return end
		if status == nil then
			frame.indicator_bl:SetBackdropColor(0,0,0,0)
		else
			frame.indicator_bl:SetBackdropColor(status.color.r, status.color.g, status.color.b, status.color.a or 1)
		end
	end)

sRaidFrames:RegisterStatusElement("indicator-br", "Indicator (Bottom Right)",
	function(self, frame, status)
		if not frame.indicator_br then return end
		if status == nil then
			frame.indicator_br:SetBackdropColor(0,0,0,0)
		else
			frame.indicator_br:SetBackdropColor(status.color.r, status.color.g, status.color.b, status.color.a or 1)
		end
	end)
